/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
* llcping.c 
*
* Program to send Test and XID frames to another Token Ring Node
* Copyright (c) Madge Networks Ltd 1999
*
*  $Id:  $
*
* Purpose:
* 	If the Madge driver was compiled with the RESPOND_TO_TEST_XID conditional
*	defined then the driver will respond to these TEST and XID packets.
*
*	It is necessary to respond to these packets if some IBM and Cabletron
*	infrastructure products are in use and they expect nodes to respond
*	to these packets.
*
*	This program was written to make sure the driver works correctly when
*	RESPOND_TO_TEST_XID is defined.
*
*	It functions in a similar manner to ping but at the MAC layer rather than
*	the IP layer.
*
*	If you are using a full LLC or SNA it is probably best to NOT define
*	RESPOND_TO_TEST_XID and to let the upper layers deal with these frames
*	properly.
*
*	Likewise, if you are using a full LLC or SNA then its probably unwise
*	to send XID frames from this program to nodes running a full LLC.
*****************************************************************************/

#include <linux/version.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <asm/types.h>
#include <linux/if_ether.h>
#include <linux/if_tr.h>
#include <net/if.h>
#include <ctype.h>

/* Local hardware address found from the IOCTL */
unsigned char hwaddr[6];

void usage()
{
	printf(
/*************************************************************/
		"Copyright (c) 1999 Madge Networks Ltd\n"
		"Usage: llcping [-t|-x] <Device> <MAC Address>\n"
		"-t	Send a test frame\n"
		"-x	Send an XID frame\n"

/* <device> is the Linux device name such as tr0, tr1 etc that
 * the test or XID frame should be sent out of.
 * The MAC address is the destination MAC address that will be
 * put in the Token Ring header.  The source MAC address will
 * be obtained by issuing an IOCTL on the device specified.
 *************************************************************/
		);
}

/* This is an arbitary constant */
#define OUR_SAP				0x0

#define LLC_TEST_XID_MASK	0xef
#define LLC_TEST			0xe3
#define LLC_XID		     	0xaf

#define TEST_DATA			"Madge Networks Ltd Linux llcping test data"
#define TEST_DATA_LEN		sizeof(TEST_DATA)

/*
 * Build a Token Ring header.
 * The format of the Token Ring header depends on the Linux version.
 * Version prior to 2.1.53 require fixed length source routing information
 * at all times. 
 */
int build_tr_header(unsigned char * buffer,unsigned char *dest)
{
	int i,srlen = 0,totallen = 0;
	struct trh_hdr * trhdr;
	struct trllc   * llchdr;
	char * data;
		
	trhdr = (struct trh_hdr *)buffer;
	trhdr->ac = 0x10;
	trhdr->fc = 0x40;
	memcpy(trhdr->saddr,hwaddr,6);
	memcpy(trhdr->daddr,dest,6);

	totallen = 14;

	trhdr->rcf = htons(0x8270);
	trhdr->saddr[0] |= TR_RII;
	srlen = 2;

#if LINUX_VERSION_CODE >= 0x020135
	totallen += srlen;
	llchdr = (struct trllc *)(buffer + totallen);
#else
	llchdr = (struct trllc *)(trhdr + 1);
	totallen += 18;
#endif

#if 0
	for(i=0;i<totallen;i++)
		printf("%02X ",buffer[i]);
	printf("\n");
#endif

	return totallen;
}

/*
 *Build a TEST packet.
 * Use the data given in testdata and testlen as the test data.
 */
int build_test_packet(unsigned char * buffer,unsigned char *dest,unsigned char *testdata,int testlen)
{
	int totallen = 0;
	struct trllc   * llchdr;
	char * data;
		
	totallen = build_tr_header(buffer,dest);

	llchdr = (struct trllc *)(buffer + totallen);

	llchdr->llc = LLC_TEST;
	llchdr->dsap = 0;
	llchdr->ssap = OUR_SAP;

	totallen += 3;

	data = (char *)(llchdr + 1);data-=5;
	
	totallen += testlen;
	
	for(;testlen>0;testlen--)
		*data++ = *testdata++;

	return totallen;
}

/*
 * Build XID packet. 
 */
int build_xid_packet(unsigned char * buffer,unsigned char * dest)
{
	int totallen = 0;
	struct trllc   * llchdr;
	char * data;

	totallen = build_tr_header(buffer,dest);

	llchdr = (struct trllc *)(buffer + totallen);

	llchdr->llc = LLC_XID;
	llchdr->dsap = 0;
	llchdr->ssap = OUR_SAP;

	totallen += 3;

	data = (char *)(llchdr + 1);data-=5;
	totallen +=3;

	data[0] = 0x81;
	data[1] = 0x01;
	data[2] = 0;

	return totallen;
}

/* Conversion functions */

char *MAC_to_string(char *str,unsigned char * macaddr)
{
	sprintf(str,"%02X%02X%02X%02X%02X%02X",
			macaddr[0],
			macaddr[1],
			macaddr[2],
			macaddr[3],			
			macaddr[4],
			macaddr[5]);
	return str;
}

unsigned char * string_to_MAC(unsigned char *macaddr,char *str)
{
	char *p;
	int	high,low,j;

	if (!str)
		return NULL;
	if (*str == '\0')
		return NULL;
		
	p = str;

	for(j=0;j<TR_ALEN;j++)
	{
		if (!isxdigit(*p))
		{
			return NULL;
		}
				
		high = 	(*p >= '0' && *p <= '9') ? *p - '0' :
				((*p >= 'A' && *p <= 'F') ? *p - 'A' + 10 : *p - 'a' + 10);
				
		p++;
				
		if (!isxdigit(*p))
		{
			return NULL;
		}
				
		low = 	(*p >= '0' && *p <= '9') ? *p - '0' :
				((*p >= 'A' && *p <= 'F') ? *p - 'A' + 10 : *p - 'a' + 10);
			
		p++;
				
		macaddr[j] = ((high << 4) | low) & 0xFF;
	} /* for(j) */

	return macaddr;
}

#define TR (*trhdr)
#define LLC (*llchdr)
int get_pointers(unsigned char * buffer,struct trh_hdr **trhdr,struct trllc **llchdr)
{
#if LINUX_VERSION_CODE >= 0x020135
	int srlen;
	__u16 rcf = ntohs(TR->rcf);
	
	TR = (struct trh_hdr *)buffer;
	srlen = (rcf & TR_RCF_LEN_MASK) >> 8;
	
	/* Check the sr length */
	if (srlen > 18 || srlen < 2)
	{
		LLC = (struct trllc *)(buffer + 14);
		return;
	}
	
	/* Check the reserved bits */
	if ((rcf & 0xF) != 0)
	{
		LLC = (struct trllc *)(buffer + 14);
		return;		
	}

	LLC = (struct trllc *)(buffer + srlen + 14);	
#else
	TR = (struct trh_hdr *)buffer;
	LLC = (struct trllc *)(TR + 1);
#endif
	return ((int)LLC) - ((int)buffer) + 3;
}
#undef TR
#undef LLC


int main(int argc,char **argv)
{
	int s;
	int len;
	int test = 1;
	struct sockaddr addr,sa;
	struct ifreq ifr;
	unsigned char Buffer[5000];
	unsigned char dest[6];
	char * device,*macstr;
	char TmpBuf1[14];
	char TmpBuf2[14];
	struct trh_hdr * trhdr;
	struct trllc   * llchdr;
	
	if (argc < 3)
	{
		usage();
		return 1;
	}

	if (getuid() != 0)
	{
		printf("You must be root to run this program\n");
		return 1;
	}

	if (*argv[1] == '-')
	{
		switch(argv[1][1])
		{
			case 't':
				test = 1;
				break;
			case 'x':
				test = 0;
				break;
			default:
				printf("Unknown option - %s\n",argv[1]);
				return 1;
		}
		device = argv[2];
		macstr = argv[3];
	}
	else
	{
		device = argv[1];
		macstr = argv[2];
	}
	
	if (NULL == string_to_MAC(dest,macstr))
	{
		printf("Couldn't parse destination MAC address\n");
		return 1;
	}

	if ((s = socket(AF_INET,SOCK_PACKET,htons(ETH_P_ALL))) < 0)
	{
		perror("socket call failed");
		return 1;
	}

	strcpy(sa.sa_data,device);
	strcpy(ifr.ifr_name,device);

	if (ioctl(s,SIOCGIFHWADDR,&ifr) < 0)
	{
		perror("Cannot get hardware address");
		close(s);
		return 1;
	}

	printf("Sending %s frame from %s (%s) to %s\n",
		test == 1 ? "test" : "XID",
		MAC_to_string(TmpBuf1,ifr.ifr_hwaddr.sa_data),device,
		MAC_to_string(TmpBuf2,dest));

	memcpy(hwaddr,ifr.ifr_hwaddr.sa_data,6);

	if (test == 1)
		len = build_test_packet(Buffer,dest,TEST_DATA,TEST_DATA_LEN);
	else
		len = build_xid_packet(Buffer,dest);

	if (sendto(s,Buffer,len,0,&sa,sizeof(sa)) < 0)
	{
		perror("sendto");
		return 1;
	}

	while(1)
	{
		int n,i;
		struct trh_hdr * trhdr;
		struct trllc * llchdr;
		len = sizeof(struct sockaddr);
		memset(Buffer,0,5000);
		n = recvfrom(s,Buffer,5000,0,&sa,&len);
		trhdr = (struct trh_hdr *)Buffer;
		trhdr->saddr[0] &= ~TR_RII; /*turn off sr bit for comparison */
		if (memcmp(trhdr->saddr,dest,6)==0)
		{
			int hlen;
			hlen = get_pointers(Buffer,&trhdr,&llchdr);
			if (llchdr->llc != UI_CMD && llchdr->dsap == OUR_SAP)
			{
				unsigned char * data;
				printf("Got Reply\n");
				printf("LLC: %02X DSAP: %02X SSAP: %02X\n",llchdr->llc,llchdr->dsap,llchdr->ssap);
				data = (unsigned char *)(llchdr + 1);data-=5;
				
				if (memcmp(data,TEST_DATA,TEST_DATA_LEN) != 0)
				{
					if (test)
						printf("Unexpected data in packet: ");
					for(i=0;i<n-hlen;i++)
						printf("%02X ",data[i]);
					printf("\n");
				} else {
					printf("Test data OK\n");
				}
				break;
			}
		}
	}

	close(s);
}
